package nl.kallestruik.darena.arenas

import net.kyori.adventure.text.Component
import nl.kallestruik.darena.arenas.world.ArenaWorld
import nl.kallestruik.darena.exceptions.ArenaStartAbortedException
import nl.kallestruik.darena.managers.ArenaManager
import nl.kallestruik.darena.managers.ConfigManager
import nl.kallestruik.darena.managers.PointsManager
import nl.kallestruik.darena.managers.TeamManager
import nl.kallestruik.darena.types.ArenaFeature
import nl.kallestruik.darena.types.arena.ProcessedArenaCheckpoint
import nl.kallestruik.darena.types.border.MCBorder
import nl.kallestruik.darena.types.countdown.AsyncCountdown
import nl.kallestruik.darena.types.countdown.AsyncCountdownWithSyncTask
import nl.kallestruik.darena.util.ArenaUtil
import nl.kallestruik.darena.util.Logger
import nl.kallestruik.darena.util.RenderUtil
import org.bukkit.Color
import org.bukkit.GameMode
import org.bukkit.World
import org.bukkit.entity.Player
import org.bukkit.plugin.java.JavaPlugin
import org.bukkit.scheduler.BukkitRunnable

class Arena(
    val config: ArenaConfig,
    private val plugin: JavaPlugin,
    private val configManager: ConfigManager,
    private val arenaManager: ArenaManager,
    private val teamManager: TeamManager,
    private val pointsManager: PointsManager,
    val world: ArenaWorld = ArenaWorld(config.name, configManager, plugin),
) {
    var session: ArenaSession? = null
    var pointsSession: PointsSession? = null

    fun start(loadingTimeOverride: Int, spectatorTimeOverride: Int, beforeStartTimeOverride: Int) {
        val loadingTime = if (loadingTimeOverride != -1) loadingTimeOverride else config.countdown.loadingTime
        val spectatorTime = if (spectatorTimeOverride != -1) spectatorTimeOverride else config.countdown.spectatorTime
        val beforeStartTime = if (beforeStartTimeOverride != -1) beforeStartTimeOverride else config.countdown.beforeStartTime

        // Create a new session
        session = ArenaSession()
        pointsSession = PointsSession(teamManager, pointsManager, this)
        // Add all participants and spectators
        session!!.originalParticipants.addAll(ArenaUtil.getPossibleParticipants())
        session!!.participants.addAll(session!!.originalParticipants)
        session!!.spectators.addAll(ArenaUtil.getPossibleSpectators())
        session!!.allPlayers.addAll(session!!.originalParticipants.union(session!!.spectators))

        session!!.currentCountdown = AsyncCountdownWithSyncTask(plugin, loadingTime, "Teleporting in", { //TODO: Make the subtitle configurable
            // Reset the world
            world.reset()
        }, {
            session!!.processSpawnRules(config.spawns, config.spawnRules, config.loadouts)
            session!!.processSpawnPools(config.spawnPools)
            session!!.processCheckpoints(config.checkpoints)

            startBorders()

            // Spawn everyone as a spectator so they can explore the map.
            for (player in session!!.allPlayers) {
                player.gameMode = GameMode.SPECTATOR
            }

            val spectatorSpawnRule = session!!.spawnRules["spectator"] ?: throw ArenaStartAbortedException("Missing spawn rule \"spectator\". This rule is required.")

            spectatorSpawnRule.spawnPlayers(session!!.allPlayers, world)

            session!!.currentCountdown = AsyncCountdown(plugin, spectatorTime, "Starting in", alwaysAsActionBar = true) { //TODO: Make the subtitle configurable
                for (participant in session!!.participants) {
                    if (config.features.contains(ArenaFeature.GAMEMODE_ADVENTURE))
                        participant.gameMode = GameMode.ADVENTURE
                    else
                        participant.gameMode = GameMode.SURVIVAL
                }

                val initialSpawnRule = session!!.spawnRules["initial"] ?: throw ArenaStartAbortedException("Missing spawn rule \"initial\". This rule is required.")
                initialSpawnRule.spawnPlayers(session!!.participants, world)

                // Mark the game as in progress. This makes the event listeners all trigger from this point onwards.
                session!!.isInProgress = true
                session!!.currentCountdown = AsyncCountdown(plugin, beforeStartTime, "Beginning in", alwaysAsTitle = true) { //TODO: Make the subtitle configurable

                    config.blockRemoveZone?.removeBlocks(world)

                    startTimers()
                }.start(session!!.allPlayers)
            }.start(session!!.allPlayers)
        }).start(session!!.allPlayers)
    }

    fun startTimers() {
        session!!.startTime = System.currentTimeMillis()
        if (config.endConditions.time != -1) {
            session!!.timeLeft = config.endConditions.time
            session!!.endTimeTask = object : BukkitRunnable() {
                override fun run() {
                    arenaManager.end()
                }
            }.runTaskLater(plugin, config.endConditions.time * 20L)
        }

        session!!.timerTask = object  : BukkitRunnable() {
            override fun run() {
                session!!.timeLeft--
                updateActionbar()
            }
        }.runTaskTimer(plugin, 20, 20)
    }

    fun startBorders() {
        config.borders.vertical.getOrNull(0)?.execute(MCBorder(world), plugin)
        config.borders.top.getOrNull(0)?.execute(session!!.borderTop, plugin)
        config.borders.bottom.getOrNull(0)?.execute(session!!.borderBottom, plugin)

        if (hasFeature(ArenaFeature.HORIZONTAL_BORDERS)) {
            session!!.borderTask = BorderRunnable(world.world, session!!).runTaskTimer(plugin, 0, 10)
        }
    }

    fun updateActionbar() {
        var actionBar = Component.empty()
        if (hasFeature(ArenaFeature.SHOW_TIMER))
            actionBar = actionBar.append(Component
                .text("Time left: ${ArenaUtil.formatTime(session!!.timeLeft)}"))
        val allPlayers = session!!.originalParticipants.union(session!!.spectators)
        for (player in allPlayers) {
            player.sendActionBar(actionBar)
        }
    }

    fun end() {
        for (player in session!!.originalParticipants) {
            ArenaUtil.clearPlayer(player)
            player.gameMode = GameMode.SURVIVAL
            session!!.spectators.remove(player)
        }

        world.empty()
        if (pointsSession != null)
            pointsSession!!.end()

        session?.timerTask?.cancel()
        session?.endTimeTask?.cancel()
        session?.currentCountdown?.cancel()
        session?.borderTask?.cancel()
    }

    fun hasFeature(feature: ArenaFeature) = config.features.contains(feature)

    fun sendMessage(component: Component) {
        for (player in session!!.allPlayers) {
            player.sendMessage(component)
        }
    }

    fun makeSpectator(player: Player) {
        Logger.trace(Arena::class, "makeSpectator(player: ${player.name})")
        player.gameMode = GameMode.SPECTATOR
        session!!.participants.remove(player)
        session!!.spectators.add(player)
    }

    fun respawn(player: Player) {
        val lastCheckpoint: ProcessedArenaCheckpoint? = session!!.lastCheckpoint[player.uniqueId]
        val spawnRule = if (lastCheckpoint != null)
            lastCheckpoint.spawnRule!!
        else
            session!!.spawnRules["initial"]!!

        spawnRule.spawnPlayer(player, world)
    }

    private class BorderRunnable(
        private val world: World,
        private val session: ArenaSession
    ): BukkitRunnable() {

        override fun run() {
            RenderUtil.drawXZPlane(world,
                -100, -100,
                100, 100,
                session.borderTop.height, Color.RED)

            RenderUtil.drawXZPlane(world,
                -100, -100,
                100, 100,
                session.borderBottom.height, Color.RED)
        }
    }
}
