package nl.kallestruik.darena.arenas

import net.kyori.adventure.text.Component
import net.kyori.adventure.text.format.TextColor
import nl.kallestruik.darena.exceptions.TeamNotFoundException
import nl.kallestruik.darena.managers.PointsManager
import nl.kallestruik.darena.managers.TeamManager
import nl.kallestruik.darena.types.Team
import nl.kallestruik.darena.types.arena.ProcessedArenaCheckpoint
import nl.kallestruik.darena.types.arena.ProcessedArenaSpawnPool
import nl.kallestruik.darena.util.Logger
import org.bukkit.entity.Player
import java.util.*
import kotlin.math.min

class PointsSession(
    private val teamManager: TeamManager,
    private val pointsManager: PointsManager,
    private val arena: Arena,
) {
    private val checkpointCompletionCount: MutableMap<String, Int> = mutableMapOf()
    private var kills: Int = 0
    // Team -> {reason -> amount}
    private val points: MutableMap<Team, MutableMap<String, Int>> = HashMap()

    @Throws(TeamNotFoundException::class)
    fun givePointsToPlayer(uuid: UUID, amount: Int, reason: String) {
        Logger.trace(PointsSession::class, "givePointsToPlayer(uuid: $uuid, amount: $amount, reason: $reason)")
        Logger.debug(PointsSession::class, "Giving $amount points to player with uuid $uuid for reason $reason")
        val team = teamManager.getTeamForPlayer(uuid) ?: return

        givePointsToTeam(team, amount, reason)
    }

    fun givePointsToTeam(team: Team, amount: Int, reason: String) {
        Logger.trace(PointsSession::class, "givePointsToTeam(team: $team, amount: $amount, reason: $reason)")
        Logger.debug(PointsSession::class, "Giving $amount points to team ${team.name} for reason $reason")
        val pointsMap = points.getOrDefault(team, mutableMapOf())
        pointsMap[reason] = (pointsMap[reason] ?: 0) + amount
        points[team] = pointsMap
    }

    fun giveCheckpointPoints(player: Player, checkpoint: ProcessedArenaCheckpoint) {
        Logger.debug(PointsSession::class, "Giving checkpoint points to player ${player.name} for reaching checkpoint ${checkpoint.label}")
        val checkpointPointList = arena.config.points.checkpoints[checkpoint.label] ?: return
        val checkpointPoints = checkpointPointList[min(checkpointCompletionCount[checkpoint.label] ?: 0, checkpointPointList.size - 1)]
        givePointsToPlayer(player.uniqueId, checkpointPoints, "Reaching checkpoint") //TODO: Make these messages configurable
        checkpointCompletionCount[checkpoint.label] = (checkpointCompletionCount[checkpoint.label] ?: 0) + 1
    }

    fun giveKillPoints(killer: Player) {
        val killPointList = arena.config.points.kill
        val killPoints = killPointList[min(kills, killPointList.size -1 )]
        givePointsToPlayer(killer.uniqueId, killPoints, "Kill")//TODO: Make these messages configurable
        kills++
    }

    fun giveLMSPoints() {
        var lmsCounter = 0
        // First award points to the players still alive
        val lmsPoints = arena.config.points.lastManStanding
        for (player in arena.session!!.participants) {
            givePointsToPlayer(player.uniqueId, lmsPoints[min(lmsCounter, lmsPoints.size - 1)], "Placement") //TODO: Make these messages configurable
            lmsCounter++
        }
        for (uuid in arena.session!!.deaths.reversed()) {
            givePointsToPlayer(uuid, lmsPoints[min(lmsCounter, lmsPoints.size - 1)], "Placement") //TODO: Make these messages configurable
            lmsCounter++
        }
    }

    fun end() {
        giveLMSPoints()

        val map = mutableMapOf<Team, Pair<Int, Component>>()

        for (entry in points) {
            val team = entry.key;
            var detailComponent = Component.text("${team.name} - Points\n")
            var totalPoints = 0
            for (reason in entry.value) {
                if (reason.value <= 0)
                    continue

                totalPoints += reason.value
                detailComponent = detailComponent.append(Component.text("\n[+${reason.value}] ${reason.key}"))
            }
            if (totalPoints > 0)
                map[team] = Pair(totalPoints, detailComponent)
        }

        val sortedMap = map.toList().sortedByDescending { (_, v) -> v.first }.toMap()

        //TODO: Make these message configurable
        arena.sendMessage(Component.text("============[ Summary ]============"))
        for (entry in sortedMap) {
            // Give points to the team.
            pointsManager.givePointsToTeam(entry.key, entry.value.first)
            arena.sendMessage(Component.empty()
                .append(teamManager.getTeamComponent(entry.key))
                .append(Component
                    .text(": "))
                .append(Component
                    .text(pointsManager.getPointsForTeam(entry.key)))
                .append(Component
                    .text(" [+${entry.value.first}]")
                    .hoverEvent(entry.value.second.asHoverEvent())
                    .color(TextColor.color(0x55FFFF)))
            )
        }
    }

    fun givePoolPoints(player: Player, spawnPool: ProcessedArenaSpawnPool, checkpointNumber: Int) {
        Logger.debug(PointsSession::class, "Giving pool points to player ${player.name} for reaching checkpoint $checkpointNumber in the pool ${spawnPool.label}")
        val poolPointList = (arena.config.points.pools[spawnPool.label] ?: return)[checkpointNumber - 1]
        val poolPoints = poolPointList[min(arena.session!!.finishedPerSection[checkpointNumber] ?: 0, poolPointList.size - 1)]
        givePointsToPlayer(player.uniqueId, poolPoints, "Reaching checkpoint") //TODO: Make these messages configurable
        arena.session!!.finishedPerSection[checkpointNumber] = (arena.session!!.finishedPerSection[checkpointNumber] ?: 0) + 1
    }


}