package nl.kallestruik.darena.commands

import co.aikar.commands.BaseCommand
import co.aikar.commands.CommandHelp
import co.aikar.commands.annotation.*
import net.kyori.adventure.text.Component
import net.kyori.adventure.text.event.ClickEvent
import net.kyori.adventure.text.format.NamedTextColor
import net.kyori.adventure.text.format.TextColor
import nl.kallestruik.darena.arenas.Arena
import nl.kallestruik.darena.exceptions.ArenaCreationException
import nl.kallestruik.darena.exceptions.ArenaEditException
import nl.kallestruik.darena.exceptions.TeamExistsException
import nl.kallestruik.darena.managers.*
import nl.kallestruik.darena.types.EditSetting
import nl.kallestruik.darena.types.Reloadable
import nl.kallestruik.darena.types.Team
import nl.kallestruik.darena.util.Logger
import org.bukkit.ChatColor
import org.bukkit.command.CommandSender
import org.bukkit.entity.Player

@CommandAlias("darena|arena")
class CommandDArena(
    private val arenaManager: ArenaManager,
    private val editManager: EditManager,
    private val pointsManager: PointsManager,
    private val teamManager: TeamManager,
    private val gameManager: GameManager,
): BaseCommand() {

    @HelpCommand
    fun doHelp(sender: CommandSender, help: CommandHelp) {
        Logger.trace(CommandDArena::class, "doHelp(sender: $sender, help: $help)")
        help.showHelp()
    }

    @Subcommand("start|s")
    @Description("Start an arena")
    @CommandPermission("darena.start")
    fun onStart(sender: CommandSender, arena: Arena, @Default("-1") loadingTimeOverride: Int, @Default("-1") spectatorTimeOverride: Int, @Default("-1") beforeStartTimeOverride: Int) {
        Logger.trace(CommandDArena::class, "onStart(sender: $sender)")
        arenaManager.start(arena, loadingTimeOverride, spectatorTimeOverride, beforeStartTimeOverride)
    }

    @Subcommand("end")
    @Description("End the current arena")
    @CommandPermission("darena.end")
    fun onEnd(sender: CommandSender) {
        Logger.trace(CommandDArena::class, "onEnd(sender: $sender)")
        arenaManager.end()
    }

    @Subcommand("list")
    @Description("List all arenas")
    @CommandPermission("darena.list")
    fun onList(sender: CommandSender) {
        Logger.trace(CommandDArena::class, "onList(sender: $sender)")
        sender.sendMessage(Component.text("============[ Arenas ]============"))
        for (gameEntry in gameManager.games) {
            sender.sendMessage(Component
                .text("${gameEntry.key}:")
                .hoverEvent(Component.text(gameEntry.value.description))
            )
            for (arena in gameEntry.value.arenas) {
                sender.sendMessage(Component.empty()
                    .append(Component
                        .text("  ${arena.config.name} "))
                    .append(Component
                        .text("[start]")
                        .clickEvent(ClickEvent.runCommand("/darena start ${arena.config.name}"))
                        .color(NamedTextColor.GREEN))
                    .append(Component
                        .text(" "))
                    .append(Component
                        .text("[edit]")
                        .clickEvent(ClickEvent.runCommand("/darena edit load ${arena.config.name}"))
                        .color(NamedTextColor.AQUA))
                )
            }
        }
    }

    @Subcommand("leaderboard|top")
    @CommandAlias("leaderboard")
    @Description("Show the points leaderboard")
    @CommandPermission("darena.leaderboard")
    fun onLeaderboard(sender: CommandSender) {
        Logger.trace(CommandDArena::class, "onLeaderboard(sender: $sender)")
        pointsManager.sendLeaderboard(sender)
    }

    @Subcommand("reload")
    @Description("Reload the config")
    @CommandPermission("darena.reload")
    fun onReload(sender: CommandSender, reloadable: Reloadable) {
        Logger.trace(CommandDArena::class, "onReload(sender: $sender, reloadable: $reloadable)")
        reloadable.reload()
        sender.sendMessage("${ChatColor.YELLOW}[?] Successfully reloaded")
    }

    @Subcommand("new")
    @Description("Create a new arena")
    @CommandPermission("darena.new")
    fun onNew(sender: CommandSender, name: String) {
        Logger.trace(CommandDArena::class, "onNew(sender: $sender, name: $name)")
        try {
            arenaManager.createArena(name)
            sender.sendMessage("${ChatColor.YELLOW}[?] Arena $name created successfully. You can edit it with /darena edit load $name")
        } catch (e: ArenaCreationException) {
            sender.sendMessage("${ChatColor.RED}[!] ${e.message}}")
        }
    }

    @Subcommand("edit")
    inner class EditCommand: BaseCommand() {

        @Subcommand("load")
        @Description("Load and teleport you to the world of an arena for editing.")
        @CommandPermission("darena.edit.load")
        fun onLoad(player: Player, arena: Arena) {
            Logger.trace(EditCommand::class, "onLoad(player: $player, arena: ${arena.config.name})")
            try {
                editManager.load(arena)
                editManager.tp(player)
            } catch (e: ArenaEditException) {
                player.sendMessage("${ChatColor.RED}[!] ${e.message}")
            }
        }

        @Subcommand("save")
        @Description("Save the world currently being edited.")
        @CommandPermission("darena.edit.load")
        fun onSave(sender: CommandSender) {
            Logger.trace(EditCommand::class, "onSave(sender: $sender)")
            try {
                editManager.save()
            } catch (e: ArenaEditException) {
                sender.sendMessage("${ChatColor.RED}[!] ${e.message}")
            }
        }

        @Subcommand("tp")
        @Description("Teleport into the world currently being edited.")
        @CommandPermission("darena.edit.tp")
        fun onTp(player: Player) {
            Logger.trace(EditCommand::class, "onTp(player: $player)")
            try {
                editManager.tp(player)
            } catch (e: ArenaEditException) {
                player.sendMessage("${ChatColor.RED}[!] ${e.message}")
            }
        }

        @Subcommand("finish")
        @Description("Save and stop editing the world currently being edited.")
        @CommandPermission("darena.edit.finish")
        fun onFinish(sender: CommandSender) {
            Logger.trace(EditCommand::class, "onFinish(sender: $sender)")
            try {
                editManager.finish()
            } catch (e: ArenaEditException) {
                sender.sendMessage("${ChatColor.RED}[!] ${e.message}")
            }
        }

        @Subcommand("config")
        @Description("Load and teleport you to the world of an arena for editing.")
        @CommandPermission("darena.edit.load")
        fun onConfig(player: Player, path: String) {
            Logger.trace(EditCommand::class, "onConfig(player: $player, path: $path)")
            player.sendMessage("This feature is currently disabled.")
//            try {
//                val info = editManager.getConfigValue(path)
//                player.sendMessage("------------------")
//                player.sendMessage("Name: ${info.name}")
//                player.sendMessage("Value: ${info.value}")
//                player.sendMessage("Description: ${info.description}")
//                player.sendMessage("------------------")
//            } catch (e: ArenaEditException) {
//                player.sendMessage("${ChatColor.RED}[!] ${e.message}")
//            }
        }

        @Subcommand("visualize")
        inner class VisualizeCommand: BaseCommand() {
            @Subcommand("spawns")
            @Description("Toggle the visibility of all spawns")
            @CommandPermission("darena.edit.visualize.spawns")
            fun onSpawns(player: Player) {
                editManager.toggleSpawnVisibility()
            }

            @Subcommand("checkpoints")
            @Description("Toggle the visibility of all checkpoints")
            @CommandPermission("darena.edit.visualize.checkpoints")
            fun onCheckpoints(player: Player) {
                editManager.toggleCheckpointVisibility()
            }
        }

        @Subcommand("toggle")
        inner class ToggleCommand: BaseCommand() {
            @Subcommand("spawnplacing")
            @Description("Toogle placing spawns with armor stands")
            @CommandPermission("darena.edit.toggle.spawnplacing")
            fun onSpawnPlacing(player: Player) {
                val newValue = editManager.toggleFeature(player, EditSetting.SPAWN_PLACING)
                player.sendMessage("Spawn placing is now: $newValue")
            }
        }

    }

    @Subcommand("team")
    inner class TeamCommand: BaseCommand() {

        @Subcommand("create")
        @Description("Create a team.")
        @CommandPermission("darena.team.create")
        fun onCreate(sender: CommandSender, teamName: String) {
            Logger.trace(TeamCommand::class, "onCreate(sender: $sender, teamName: $teamName)")
            try {
                teamManager.createTeam(teamName)
                sender.sendMessage("${ChatColor.YELLOW}[?] Team $teamName was successfully created.")
            } catch (e: TeamExistsException) {
                sender.sendMessage("${ChatColor.RED}[!] ${e.message}")
            }
        }

        @Subcommand("join|add-player")
        @Description("Add a player to a team.")
        @CommandPermission("darena.team.join")
        @CommandCompletion("@team @players")
        fun onAddPlayer(sender: CommandSender, team: Team, player: Player) {
            Logger.trace(TeamCommand::class, "onAddPlayer(sender: $sender, team: $team, player: $player)")
            if (teamManager.getTeamForPlayer(player.uniqueId) != null) {
                sender.sendMessage("${ChatColor.RED}[!] Player is already in a team!")
                return
            }

            teamManager.addPlayerToTeam(team, player.uniqueId)
            sender.sendMessage("${ChatColor.YELLOW}[?] Player was successfully added to team.")
        }

        @Subcommand("leave|remove-player")
        @Description("Remove a player from a team.")
        @CommandPermission("darena.team.leave")
        @CommandCompletion("@players")
        fun onRemovePlayer(sender: CommandSender, player: Player) {
            Logger.trace(TeamCommand::class, "onRemovePlayer(sender: $sender, player: $player)")
            if (teamManager.getTeamForPlayer(player.uniqueId) == null) {
                sender.sendMessage("${ChatColor.RED}[!] Player is not in a team!")
                return
            }

            teamManager.removePlayerFromTeam(player.uniqueId)
            sender.sendMessage("${ChatColor.YELLOW}[?] Player was successfully removed from their team.")
        }

        @Subcommand("remove")
        @Description("Remove a team.")
        @CommandPermission("darena.team.remove")
        fun onRemove(sender: CommandSender, team: Team) {
            Logger.trace(TeamCommand::class, "onRemove(sender: $sender, team: $team)")
            teamManager.removeTeam(team)
            sender.sendMessage("${ChatColor.YELLOW}[?] Team ${team.name} was successfully removed.")
        }

        @Subcommand("remove-all")
        @Description("Remove all teams.")
        @CommandPermission("darena.team.remove-all")
        fun onRemoveAll(sender: CommandSender) {
            Logger.trace(TeamCommand::class, "onRemoveAll(sender: $sender)")
            val count = teamManager.removeAllTeams()
            pointsManager.clear()
            sender.sendMessage("${ChatColor.YELLOW}[?] Removed $count teams.")
        }

        @Subcommand("auto")
        @Description("Remove all current teams and enable/disable automatic team mode.")
        @CommandPermission("darena.team.auto")
        fun onAuto(sender: CommandSender, enable: Boolean) {
            Logger.trace(TeamCommand::class, "onAuto(sender: $sender, enable: $enable)")
            teamManager.autoTeams = enable
            teamManager.removeAllTeams()
            if (enable)
                teamManager.autoTeam()

            sender.sendMessage("${ChatColor.YELLOW}[?] Automatic team creation is now ${if (enable) ("${ChatColor.GREEN}enabled") else "${ChatColor.RED}disabled"}")
        }

        @Subcommand("option")
        @Description("Change team options.")
        @CommandPermission("darena.team.option")
        fun onOption(sender: CommandSender, team: Team, @Values("@teamOptions") option: String, value: String) {
            Logger.trace(TeamCommand::class, "onOption(sender: $sender, option: $option, value: $value)")
            try {
                when (option.lowercase()) {
                    "color" -> team.color = TextColor.fromCSSHexString(value)!!
                    "prefix" -> team.prefix = value.replace("\"", "")
                    else -> sender.sendMessage("${ChatColor.RED}[!] Option $option does not exist!")
                }
            } catch (e: Exception) {
                sender.sendMessage("${ChatColor.RED}[!] Something went wrong while trying to update the value of $option!")
            }
        }

        @Subcommand("list")
        @Description("Get a list of all teams and their members.")
        @CommandPermission("darena.team.list")
        fun onList(sender: CommandSender) {
            Logger.trace(TeamCommand::class, "onList(sender: $sender)")
            //TODO: Make these configurable
            sender.sendMessage("============[ Teams ]============")
            for (teamName in teamManager.getAllTeamNames()) {
                val team = teamManager.getTeamByName(teamName)

                sender.sendMessage(teamManager.getTeamComponent(team))
            }
        }
    }

    @Subcommand("points")
    inner class PointsCommand: BaseCommand() {

        @Subcommand("add")
        @Description("Add points to a team.")
        @CommandPermission("darena.points.add")
        fun onAdd(sender: CommandSender, team: Team, amount: Int) {
            Logger.trace(PointsCommand::class, "onAdd(sender: $sender, team: ${team.name}, amount: $amount)")
            pointsManager.givePointsToTeam(team, amount)
            sender.sendMessage("${ChatColor.YELLOW}[?] Added $amount points to the score of ${team.name}")
        }

        @Subcommand("remove")
        @Description("Remove points from a team.")
        @CommandPermission("darena.points.remove")
        fun onRemove(sender: CommandSender, team: Team, amount: Int) {
            Logger.trace(PointsCommand::class, "onRemove(sender: $sender, team: ${team.name}, amount: $amount)")
            pointsManager.givePointsToTeam(team, -amount)
            sender.sendMessage("${ChatColor.YELLOW}[?] Removed $amount points from the score of ${team.name}")
        }

        @Subcommand("reset")
        @Description("Reset everyones score.")
        @CommandPermission("darena.points.reset")
        fun onReset(sender: CommandSender) {
            Logger.trace(PointsCommand::class, "onReset(sender: $sender)")
            pointsManager.clear()
            sender.sendMessage("${ChatColor.YELLOW}[?] Reset everyone's score")
        }
    }
}