package nl.kallestruik.darena.managers

import net.kyori.adventure.text.Component
import net.kyori.adventure.text.format.NamedTextColor
import nl.kallestruik.darena.exceptions.TeamNotFoundException
import nl.kallestruik.darena.types.Team
import nl.kallestruik.darena.util.ConfigHelper
import nl.kallestruik.darena.util.Logger
import org.bukkit.Bukkit
import org.bukkit.command.CommandSender
import org.bukkit.configuration.file.YamlConfiguration
import java.io.File
import java.util.*

class PointsManager(
    val teamManager: TeamManager,
    val pointsFile: File
) {
    private val points: MutableMap<Team, Int> = HashMap()

    fun load() {
        val config = ConfigHelper.getOrCreateConfig(pointsFile, "template/points.yml")
        if (config.isConfigurationSection("teams")) {
            for (teamName in config.getConfigurationSection("teams")!!.getKeys(false)) {
                try {
                    val team = teamManager.getTeamByName(teamName)
                    points[team] = config.getInt("teams.$teamName")
                } catch (e: TeamNotFoundException) {
                    Logger.warn(PointsManager::class, "Team $teamName is present in points.yml but does not exist ignoring.")
                }
            }
        }
        updateTeamSuffixForAll()
    }

    fun save() {
        val config = YamlConfiguration()
        val teamsSection = config.createSection("teams")
        for (entry in points) {
            teamsSection.set(entry.key.name, entry.value)
        }

        config.save(pointsFile)
    }

    @Throws(TeamNotFoundException::class)
    fun givePointsToPlayer(uuid: UUID, amount: Int) {
        val team = teamManager.getTeamForPlayer(uuid) ?: throw TeamNotFoundException()

        givePointsToTeam(team, amount)
    }

    fun givePointsToTeam(team: Team, amount: Int) {
        points[team] = (points[team] ?: 0) + amount
        updateTeamSuffix(team)
    }

    fun updateTeamSuffix(team: Team) {
        team.mcTeam.suffix(
            Component
                .text(" [${getPointsForTeam(team)}]")
                .color(NamedTextColor.AQUA)
        )
    }

    fun updateTeamSuffixForAll() {
        for (team in teamManager.getAllTeams()) {
            updateTeamSuffix(team)
        }
    }

    @Throws(TeamNotFoundException::class)
    fun getPointsForPlayer(uuid: UUID): Int {
        val team = teamManager.getTeamForPlayer(uuid) ?: throw TeamNotFoundException()

        return getPointsForTeam(team)
    }

    fun getPointsForTeam(team: Team): Int {
        return points[team] ?: 0
    }

    fun sendLeaderboard(sender: CommandSender) {
        val sortedMap = points.toList().sortedByDescending { (_, v) -> v }.toMap()

        //TODO: Make these message configurable
        sender.sendMessage(Component.text("============[ Leaderboard ]============"))
        var i = 1
        for (entry in sortedMap) {
            var teamMembers: Component = Component.empty()
            for (uuid in teamManager.getTeamMembers(entry.key)) {
                val player = Bukkit.getPlayer(uuid)
                teamMembers = teamMembers.append(Component.text("${player?.name}\n"))
            }
            sender.sendMessage(
                Component.empty()
                    .append(Component.text("$i. "))
                    .append(teamManager.getTeamComponent(entry.key))
                    .append(Component.text(": "))
                    .append(Component.text(entry.value))
            )
            i++
        }
    }

    fun clear() {
        points.clear()
        updateTeamSuffixForAll()
    }
}